/**
 * This file demonstrates the ability to get info from a group.
 */

#include <stdio.h>

#include "hebi.h"
#include "util/lookup_helpers.c"
#include "util/hebi_util.h"

int main(int argc, char* argv[])
{
  // Try and get the requested group.
  HebiGroupPtr group = getGroupFromArgs(argc, argv);
  if (group == NULL)
  {
    hebiCleanup();
    return -1;
  }

  // How many modules are in the group?
  int num_modules = hebiGroupGetSize(group);

  // Create a group info object; this will be filled in during the request.
  HebiGroupInfoPtr info = hebiGroupInfoCreate(num_modules);
  if (info == NULL)
  {
    printf("Could not create info object!\n");
    hebiGroupRelease(group);
    hebiCleanup();
    return -1;
  }

  // In a loop, send requests for info to the group and wait for responses.
  long timeout_ms = 1000;
  float period_s = 0.25f;
  int i;
  for (i = 0; i < 20; i++)
  {
    if (hebiGroupRequestInfo(group, info, timeout_ms) == 0)
    {
      int mod_idx;
      for (mod_idx = 0; mod_idx < num_modules; mod_idx++)
      {
        printf("Module index: %d\n", mod_idx);
        HebiInfoPtr module_info =
          hebiGroupInfoGetModuleInfo(info, mod_idx);
        // The info object contains many fields and subobjects, depending on
        // the type of info. Here we get a subcommand of this object which
        // contains relevant info for actuators, such as name and a couple
        // values of gains.
        char buf[50];
        buf[0] = 'n';
        buf[1] = '/';
        buf[2] = 'a';
        buf[3] = 0;

        size_t str_len = 50;
        int ctrl_stg;
        float p_kp, v_kp;

        if (hebiInfoGetString(module_info, HebiInfoStringName, buf, &str_len) != HebiStatusSuccess)
        {
          buf[0] = '?';
          buf[1] = '\0';
        }

        if (hebiInfoGetFloat(module_info, HebiInfoFloatPositionKp, &p_kp) != HebiStatusSuccess)
          p_kp = NAN;
        if (hebiInfoGetFloat(module_info, HebiInfoFloatVelocityKp, &v_kp) != HebiStatusSuccess)
          v_kp = NAN;
        if (hebiInfoGetEnum(module_info, HebiInfoEnumControlStrategy, &ctrl_stg) != HebiStatusSuccess)
          ctrl_stg = -1;

        printf("Name %s | Position k_p %f | Velocity k_p %f\n | Control_strategy %d\n ",
          buf, p_kp, v_kp, ctrl_stg);
      }
    }
    else
    {
      printf("Received no info from group!\n");
    }
    hebi_sleep_ms(period_s * 1000);
  }

  // We must release the resources back to the library when done!
  hebiGroupInfoRelease(info);
  hebiGroupRelease(group);
  hebiCleanup();
  return 0;
}
