#include "kinematics_helpers.h"

/*
 * Helper function to set values in a homogenous transform.
 */
static void hebiSetMatrix4dIdentity(double* matrix) {
  matrix[0] = 1;
  matrix[1] = 0;
  matrix[2] = 0;
  matrix[3] = 0;
  matrix[4] = 0;
  matrix[5] = 1;
  matrix[6] = 0;
  matrix[7] = 0;
  matrix[8] = 0;
  matrix[9] = 0;
  matrix[10] = 1;
  matrix[11] = 0;
  matrix[12] = 0;
  matrix[13] = 0;
  matrix[14] = 0;
  matrix[15] = 1;
}

/*
 * Helper function to set values in a homogenous transform.
 */
static void hebiSetMatrix4dRotX(double* matrix, double radians) {
  matrix[0] = 1;
  matrix[1] = 0;
  matrix[2] = 0;
  matrix[4] = 0;
  matrix[5] = cos(radians);
  matrix[6] = -sin(radians);
  matrix[8] = 0;
  matrix[9] = sin(radians);
  matrix[10] = cos(radians);
}

/*
 * Helper function to set values in a homogenous transform.
 */
static void hebiSetMatrix4dTranslate(double* matrix, double x, double y, double z) {
  matrix[3] = x;
  matrix[7] = y;
  matrix[11] = z;
}

/*
 * Helper function to set values in a vector.
 */
static void hebiSetVector3d(double* vector, double x, double y, double z) {
  vector[0] = x;
  vector[1] = y;
  vector[2] = z;
}

/*
 * Helper function to set values in 3D inertia tensor.
 */
static void hebiSetSphereInertia(double* inertia, double mass, double radius)
{
  // Ixx = Iyy = Izz = 2mr^2 / 5
  // Ixy = Ixz = Iyz = 0;
  inertia[0] = inertia[1] = inertia[2] = 0.4f * mass * radius * radius;
  inertia[3] = inertia[4] = inertia[5] = 0;
}

/*
 * Helper function to set values in 3D inertia tensor. This assumes a thin rod
 * along the x axis, giving the inertia about the center of mass.
 */
static void hebiSetRodXAxisInertia(double* inertia, double mass, double length)
{
  // Iyy = Izz = ml^2 / 12
  // Ixx = Ixy = Ixz = Iyz = 0;
  inertia[1] = inertia[2] = mass * length * length / 12.f;
  inertia[0] = inertia[3] = inertia[4] = inertia[5] = 0;
}

// Helper functions to clean up below; I'd make this private if I could ;)
HebiRobotModelElementPtr tryAdd(
  HebiRobotModelPtr robot_model,
  HebiRobotModelElementPtr existing_element,
  size_t output_index,
  HebiRobotModelElementPtr new_element,
  int32_t combine) // 1 for yes, 0 for no
{
  if (new_element == NULL)
  {
    return NULL;
  }
  HebiStatusCode res = hebiRobotModelAdd(
    robot_model, existing_element, output_index, new_element, combine);
  if (res == HebiStatusFailure)
  {
    hebiRobotModelElementRelease(new_element);
    return NULL;
  }
  return new_element;
}

int isX5(HebiActuatorType actuator_type)
{
  return (actuator_type == HebiActuatorTypeX5_1 ||
         actuator_type == HebiActuatorTypeX5_4 ||
         actuator_type == HebiActuatorTypeX5_9) ? 1 : 0;
}

int isX8(HebiActuatorType actuator_type)
{
  return (actuator_type == HebiActuatorTypeX8_3 ||
         actuator_type == HebiActuatorTypeX8_9 ||
         actuator_type == HebiActuatorTypeX8_16) ? 1 : 0;
}

HebiRobotModelElementPtr hebiRobotModelAddActuator(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiActuatorType actuator_type, int32_t combine)
{
  double com[16];
  hebiSetMatrix4dIdentity(com);

  double input_to_axis[16];
  hebiSetMatrix4dIdentity(input_to_axis);

  double mass = 0;

  HebiJointType joint_type = HebiJointTypeRotationZ;

  switch (actuator_type)
  {
    case HebiActuatorTypeX5_1:
      mass = 0.315f;
      break;
    case HebiActuatorTypeX5_4:
      mass = 0.335f;
      break;
    case HebiActuatorTypeX5_9:
      mass = 0.360f;
      break;
    case HebiActuatorTypeX8_3:
      mass = 0.460f;
      break;
    case HebiActuatorTypeX8_9:
      mass = 0.480f;
      break;
    case HebiActuatorTypeX8_16:
      mass = 0.500f;
      break;
    default:
      return NULL;
  }

  double inertia[6];
  if (isX5(actuator_type) == 1)
  {
    hebiSetMatrix4dTranslate(com, -0.0142, -0.0031, 0.0165);
    hebiSetMatrix4dTranslate(input_to_axis, 0.0, 0.0, 0.03105);
    inertia[0] = 0.00015; // XX
    inertia[1] = 0.000255; // YY
    inertia[0] = 0.000350; // ZZ
    inertia[0] = 0.0000341; // XY
    inertia[0] = 0.0000118; // XZ
    inertia[0] = 0.00000229; // YZ
  }
  else if (isX8(actuator_type) == 1)
  {
    hebiSetMatrix4dTranslate(com, -0.145, -0.0031, 0.0242);
    hebiSetMatrix4dTranslate(input_to_axis, 0.0, 0.0, 0.0451);
    hebiSetSphereInertia(inertia, mass, 0.06);
    inertia[0] = 0.000246; // XX
    inertia[1] = 0.000380; // YY
    inertia[0] = 0.000463; // ZZ
    inertia[0] = 0.0000444; // XY
    inertia[0] = 0.0000266; // XZ
    inertia[0] = 0.00000422; // YZ
  }
  else
  {
    return NULL;
  }

  ///////////////////////////////////////////////////////////////
  // Create the actuator (currently, this is X-series specific)
  ///////////////////////////////////////////////////////////////


  // Create and add the body (and handle failure cases)
  HebiRobotModelElementPtr body = hebiRobotModelElementCreateRigidBody(
    com, inertia, mass, 1, input_to_axis);
  if (tryAdd(robot_model, existing_element, output_index, body, combine) == NULL)
    return NULL;

  // Create and add the joint (and handle failure cases)
  // Note -- weird case if this fails, as we've already added the body...
  // leaves the kinematic object in an odd state.
  HebiRobotModelElementPtr joint = hebiRobotModelElementCreateJoint(
    joint_type);
  if (tryAdd(robot_model, body, 0, joint, 1) == NULL)
    return NULL;

  return joint;
}

HebiRobotModelElementPtr hebiRobotModelAddLink(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiLinkType link_type, double extension, double twist)
{
  double com[16];
  hebiSetMatrix4dIdentity(com);

  double output[16];
  hebiSetMatrix4dIdentity(output);
  const size_t num_outputs = 1;

  // Tube approx. 0.4kg / 1m; 0.03 m shorter than the total extension length
  // End brackets + hardware approx. 0.26 kg
  double mass = 0.4f * (extension - 0.03) + 0.26f;

  if (link_type != HebiLinkTypeX5_8)
    return NULL;

  // Edge of bracket to center of pipe.
  double edge_to_center = .0175f;

  // Note that this ignores the effect of the end brackets on moving the com
  // slightly off center.
  hebiSetMatrix4dTranslate(com, extension * 0.5f, 0, edge_to_center);
  hebiSetMatrix4dIdentity(output);
  hebiSetMatrix4dRotX(output, twist);
  hebiSetMatrix4dTranslate(output, extension,
                           -edge_to_center * sin(twist),
                           edge_to_center * (1 + cos(twist)));

  // Inertia: current approximation is a thin rod; will be improved in future.
  double inertia[6];
  hebiSetRodXAxisInertia(inertia, mass, extension);

  // Generic -- (try to) create and add the body
  HebiRobotModelElementPtr body =
    hebiRobotModelElementCreateRigidBody(com, inertia, mass, num_outputs, output);
  if (tryAdd(robot_model, existing_element, output_index, body, 0) == NULL)
    return NULL;
  return body;
}

HebiRobotModelElementPtr hebiRobotModelAddBracket(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiBracketType bracket_type)
{
  double com[16];
  hebiSetMatrix4dIdentity(com);

  double output[16];
  hebiSetMatrix4dIdentity(output);
  const size_t num_outputs = 1;

  double mass = 0;

  // Light bracket
  switch(bracket_type)
  {
    case HebiBracketTypeX5LightLeft:
    case HebiBracketTypeX5LightRight:
    {
      double mult = 1;
      if (bracket_type == HebiBracketTypeX5LightRight)
        mult = -1;

      mass = 0.1;

      hebiSetMatrix4dTranslate(com, 0, mult * 0.0215f, 0.02f);
      hebiSetMatrix4dRotX(output, mult * (-M_PI / 2.0f));
      hebiSetMatrix4dTranslate(output, 0, mult * .043f, 0.04f);
      break;
    }
    case HebiBracketTypeX5HeavyLeftInside:
    case HebiBracketTypeX5HeavyLeftOutside:
    case HebiBracketTypeX5HeavyRightInside:
    case HebiBracketTypeX5HeavyRightOutside:
    {
      double lr_mult = 1;
      if (bracket_type == HebiBracketTypeX5HeavyRightInside ||
          bracket_type == HebiBracketTypeX5HeavyRightOutside)
        lr_mult = -1;

      double y_dist = -0.0225; // Inside
      if (bracket_type == HebiBracketTypeX5HeavyLeftOutside ||
          bracket_type == HebiBracketTypeX5HeavyRightOutside)
        y_dist = 0.0375; // Outside

      mass = 0.215;

      hebiSetMatrix4dTranslate(com, 0, lr_mult * 0.5 * y_dist, 0.0275);
      hebiSetMatrix4dRotX(output, lr_mult * (-M_PI / 2.0));
      hebiSetMatrix4dTranslate(output, 0, lr_mult * y_dist, 0.055);
      break;
    }
    default:
      return NULL;
  }

  // Inertia: current approximation is a 6cm sphere; will be improved in future.
  double inertia[6];
  hebiSetSphereInertia(inertia, mass, 0.06);

  // Generic -- (try to) create and add the body
  HebiRobotModelElementPtr body =
    hebiRobotModelElementCreateRigidBody(com, inertia, mass, num_outputs, output);
  if (tryAdd(robot_model, existing_element, output_index, body, 0) == NULL)
    return NULL;

  return body;
}
