#pragma once

#include "hebi.h"

typedef enum HebiActuatorType
{
  HebiActuatorTypeX5_1,
  HebiActuatorTypeX5_4,
  HebiActuatorTypeX5_9,
  HebiActuatorTypeX8_3,
  HebiActuatorTypeX8_9,
  HebiActuatorTypeX8_16,
} HebiActuatorType;

typedef enum HebiLinkType
{
  HebiLinkTypeX5_8
} HebiLinkType;

typedef enum HebiBracketType
{
  HebiBracketTypeX5LightLeft,
  HebiBracketTypeX5LightRight,
  HebiBracketTypeX5HeavyLeftInside,
  HebiBracketTypeX5HeavyLeftOutside,
  HebiBracketTypeX5HeavyRightInside,
  HebiBracketTypeX5HeavyRightOutside
} HebiBracketType;

/**
 * \brief Add the specified actuator to a parent element in the robot model.
 *
 * After the addition, the robot model object manages the resources of the added
 * element.
 *
 * The added actuator is assumed to connect to an available output on an element
 * that has already been attached to the kinematic tree. That element should be
 * passed in as 'existing_element', and the index of the requested output on
 * that element should be given as 'output_index'.
 *
 * To attach the initial element to the robot model object, use NULL for the
 * 'existing_element' argument.
 *
 * NOTE: currently, only a single output is supported for each element (e.g., a
 * kinematic chain), and so the 'existing_element' and 'output_index' parameters
 * are not checked; this will be changed in an upcoming minor release so do not
 * rely on this behavior!
 *
 * \param robot_model A valid HEBI RobotModel object.
 * \param existing_element The parent element which the actuator is added to (or
 * NULL to add the initial element to the tree).
 * \param output_index The index of the requested output on the parent element on
 * on which to attach this actuator.
 * \param actuator_type The type of actuator to add to the robot model.
 * \param combine '0' to add (not combine) with the previous body, '1' to
 * combine with it.
 *
 * \returns On success, the RobotModelElementPtr corresponding to the actuator's
 * output. Otherwise, NULL is returned (e.g., the parent element's requested
 * output is invalid or already occupied).
 */
HebiRobotModelElementPtr hebiRobotModelAddActuator(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiActuatorType actuator_type, int32_t combine);

/**
 * \brief Add the specified link to a parent element in the robot model.
 *
 * After the addition, the robot model object manages the resources of the added
 * element.
 *
 * The added link is assumed to connect to an available output on an element
 * that has already been attached to the kinematic tree. That element should be
 * passed in as 'existing_element', and the index of the requested output on
 * that element should be given as 'output_index'.
 *
 * To attach the initial element to the robot model object, use NULL for the
 * 'existing_element' argument.
 *
 * NOTE: currently, only a single output is supported for each element (e.g., a
 * kinematic chain), and so the 'existing_element' and 'output_index' parameters
 * are not checked; this will be changed in an upcoming minor release so do not
 * rely on this behavior!
 *
 * \param robot_model A valid HEBI RobotModel object.
 * \param existing_element The parent element which the link is added to (or
 * NULL to add the initial element to the tree).
 * \param output_index The index of the requested output on the parent element on
 * on which to attach this link.
 * \param extension The length, from center of the previous element's output to
 * the center of the subsequent element's input, in meters.
 * \param twist The twist about the central axis of the link; see docs.hebi.us
 * for documentation of different link types.
 *
 * \returns On success, the RobotModelElementPtr corresponding to the link's
 * output. Otherwise, NULL is returned (e.g., the parent element's requested
 * output is invalid or already occupied).
 */
HebiRobotModelElementPtr hebiRobotModelAddLink(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiLinkType link_type, double extension, double twist);

/**
 * \brief Add the specified bracket to a parent element in the robot model.
 *
 * After the addition, the robot model object manages the resources of the added
 * element.
 *
 * The added bracket is assumed to connect to an available output on an element
 * that has already been attached to the kinematic tree. That element should be
 * passed in as 'existing_element', and the index of the requested output on
 * that element should be given as 'output_index'.
 *
 * To attach the initial element to the robot model object, use NULL for the
 * 'existing_element' argument.
 *
 * NOTE: currently, only a single output is supported for each element (e.g., a
 * kinematic chain), and so the 'existing_element' and 'output_index' parameters
 * are not checked; this will be changed in an upcoming minor release so do not
 * rely on this behavior!
 *
 * \param robot_model A valid HEBI RobotModel object.
 * \param existing_element The parent element which the bracket is added to (or
 * NULL to add the initial element to the tree).
 * \param output_index The index of the requested output on the parent element on
 * on which to attach this bracket.
 * \param bracket_type The type of bracket to add to the robot model.
 *
 * \returns On success, the RobotModelElementPtr corresponding to the bracket's
 * output. Otherwise, NULL is returned (e.g., the parent element's requested
 * output is invalid or already occupied).
 */
HebiRobotModelElementPtr hebiRobotModelAddBracket(
  HebiRobotModelPtr robot_model, HebiRobotModelElementPtr existing_element, size_t output_index,
  HebiBracketType bracket_type);
