/**
 * This file demonstrates the ability to command a group based on the feedback
 * it is sending.
 */

#include <stdio.h>
#include <inttypes.h>

#include "hebi.h"

////////////////////////////////////////////////////////////////////////////////
// These are kept as global variables for the purpose of example readability.
////////////////////////////////////////////////////////////////////////////////
HebiLookupPtr lookup = NULL;
HebiGroupPtr group = NULL;
HebiGroupCommandPtr group_cmd = NULL;
HebiGroupFeedbackPtr group_fbk = NULL;

////////////////////////////////////////////////////////////////////////////////
// Demonstrates getting feedback in syncronous and asyncronous manner
////////////////////////////////////////////////////////////////////////////////
int example()
{
  //////////////////////////////////////////////////////////////////////////////
  // Get the group
  //////////////////////////////////////////////////////////////////////////////
  lookup = hebiLookupCreate();
  if (lookup == NULL)
  {
    printf("createLookup() returned NULL.\n\n");
    return 1;
  }

  // Actually create the group (NOTE: change these values and the corresponding
  // buffer lengths when looking up different modules)
  char* names[3];
  int num_names = sizeof(names) / sizeof(names[0]);
  char name_1[5] = "base";
  names[0] = name_1;
  char name_2[9] = "shoulder";
  names[1] = name_2;
  char name_3[6] = "elbow";
  names[2] = name_3;

  char* families[1];
  const int num_families = sizeof(families) / sizeof(families[0]);
  char fam_name_1[7] = "family";
  families[0] = fam_name_1;

  // Test out lookup methods; use a 4 second timeout.
  long timeout_ms = 4000;

  group = hebiGroupCreateFromNames(lookup,
                                   (const char* const*) families, num_families,
                                   (const char* const*) names, num_names, timeout_ms);
  if (group == NULL)
  {
    printf("Group not found on network.\n\n");
    return 1;
  }

  // Create the group command and feedback objects
  int num_modules = hebiGroupGetSize(group);
  group_cmd = hebiGroupCommandCreate(num_modules);
  if (group_cmd == NULL)
  {
    printf("Could not create group command.\n\n");
    return 1;
  }
  group_fbk = hebiGroupFeedbackCreate(num_modules);
  if (group_fbk == NULL)
  {
    printf("Could not create group feedback.\n\n");
    return 1;
  }

  // Set command lifetime to 100ms.
  if (hebiGroupSetCommandLifetime(group, 100) != 0)
  {
    printf("Could not set command lifetime.\n\n");
    return 1;
  }

  // Control the robot at 100Hz for 30 seconds
  printf("Running demo for 30 seconds...\n\n");
  int timestep = 0;
  const int time_s = 30;
  const int frequency_hz = 100;
  if (hebiGroupSetFeedbackFrequencyHz(group, frequency_hz) != 0)
  {
    printf("Could not set feedback frequency.\n\n");
    return 1;
  }

  for (timestep = 0; timestep < time_s * frequency_hz; ++timestep)
  {
    // At each timestep, command a torque or force proportional to the
    // displacement (i,e., create a virtual spring).

    // Get the current feedback (this blocking call essentially sets the loop
    // rate based on our feedback frequency)
    hebiGroupGetNextFeedback(group, group_fbk, timeout_ms);

    // Fill in the effort value for each module
    int64_t revs; float offset;
    for (int i = 0; i < num_modules; ++i)
    {
      // Get angle (if present in feedback)
      HebiFeedbackPtr fbk = hebiGroupFeedbackGetModuleFeedback(group_fbk, i);
      if (hebiFeedbackGetHighResAngle(fbk, HebiFeedbackHighResAnglePosition, &revs, &offset) != HebiStatusSuccess)
      {
        printf("Could not get high res angle.\n\n");
        continue;
      }
      double angle = (double)revs * M_PI + offset;

      // Apply Hooke's law: F = -k * x
      // Units are, e.g., Nm / rad for rotary modules or N / m for linear
      double spring_constant = -10;
      HebiCommandPtr cmd = hebiGroupCommandGetModuleCommand(group_cmd, i);

      float effort = (float)(spring_constant * angle);
      hebiCommandSetFloat(cmd, HebiCommandFloatEffort, &effort);
    }

    hebiGroupSendCommand(group, group_cmd);
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////
// Main file is responsible for cleanup here
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  // Run example:
  int res = example();

  // Cleanup:
  if (group_cmd != NULL)
    hebiGroupCommandRelease(group_cmd);
  if (group_fbk != NULL)
    hebiGroupFeedbackRelease(group_fbk);
  if (group != NULL)
    hebiGroupRelease(group);
  if (lookup != NULL)
    hebiLookupRelease(lookup);

  hebiCleanup();

  return res;
}
